import asyncio
import socket
import ssl
import logging

logger = logging.getLogger(name="general")


def get_ssl_context(
        enable=False, cafile=None, client_crt=None, client_key=None):
    if not enable:
        return None
    ssl_context = ssl.create_default_context(
        ssl.Purpose.SERVER_AUTH, cafile=cafile)
    ssl_context.check_hostname = False
    ssl_context.load_cert_chain(client_crt, client_key)
    return ssl_context


async def get_message_from_queue(queue):
    message = None
    try:
        message = await asyncio.wait_for(queue.get(), timeout=1)
        queue.task_done()
    except asyncio.TimeoutError:
        pass
    return message


async def get_connection(host, port, ssl_context=None):
    try:
        reader, writer = await asyncio.open_connection(
            host=host, port=port,
            ssl=ssl_context,
            family=socket.AF_INET)
        return reader, writer
    except ConnectionError:
        logger.debug(
            ("cant connect to {}:{}, "
             "going to try again").format(
                host, port))
        return None


async def send_message(reader, writer, message):
    if reader.at_eof():
        logger.error("Server disconnected while transfer")
        return True, message
    try:
        writer.write(message.encode())
        await writer.drain()
    except Exception as e:
        logger.error("Exception: {}".format(e))
        return True, message
    return False, None


async def logstash_connection(queue, state, loop, config):
    ssl_context = get_ssl_context(
        enable=config['ssl']['enable'], cafile=config['ssl']['cafile'],
        client_crt=config['ssl']['client_crt'],
        client_key=config['ssl']['client_key']
    )

    message = None
    need_reconnect = True
    conn = None
    host = config["connection"]['host']
    port = config["connection"]['port']
    while (state.need_shutdown is False) or queue.qsize() > 0:
        if need_reconnect:
            logger.info("connecting to server")
            conn = await get_connection(host, port, ssl_context=ssl_context)
        if not conn:
            logger.info("reconnecting")
            await asyncio.sleep(1)
            need_reconnect = True
            continue
        need_reconnect = False

        if not message:
            message = await get_message_from_queue(queue)

        if message:
            need_reconnect, message = await send_message(*conn, message)
    reader, writer = conn
    writer.close()



from lsshipper.common.config import prepare_config
import asyncio
from .file_handler import FileHandler
import signal
from functools import partial
import logging
import logging.config
from .common.state import State

signal_times = 0
logger = logging.getLogger('general')


def got_int_signal(state, signum, frame):
    global signal_times
    signal_times += 1
    state.shutdown()
    if signal_times > 1:
        state.loop.close()


def main():
    config = prepare_config()
    loop = asyncio.get_event_loop()
    state = State(loop)
    shipper = FileHandler(loop=loop, state=state, config=config)
    task = asyncio.ensure_future(shipper.start())
    signal.signal(signal.SIGINT, partial(got_int_signal, state))

    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt as e:
        logger.info("process stopped by keyboard interrupt")



import os
import aiofiles


class FileReader(object):
    def __init__(self, filename, offset=0, sep=b'\r\n'):
        self.name = os.path.abspath(filename)
        self.offset = 0
        self.sep = sep  # line separator

    async def get_line(self, chunk_size=4096):
        async with aiofiles.open(self.name, 'r+b') as f:
            await f.seek(self.offset)
            tmp_line = b""
            cont = True
            while cont:
                chunk = await f.read(chunk_size)
                if not chunk:
                    yield (None, self.offset)
                    break
                if chunk[-1] == 0:
                    cont = False
                if len(tmp_line):
                    chunk = tmp_line + chunk
                    tmp_line = b''
                while chunk:
                    line, sep, chunk = chunk.partition(self.sep)
                    if sep is self.sep:
                        self.offset = self.offset + len(line) + len(self.sep)
                        yield (line, self.offset)
                    if sep is b'':
                        tmp_line = tmp_line + line
                if not cont:
                    yield (None, self.offset)



import pytest
from lsshipper.reader_aio import FileReader

file1 = "tests/lsshipper/test_files/test1.txt"
file2 = "tests/lsshipper/test_files/test2.txt"


@pytest.mark.asyncio
@pytest.mark.parametrize("chunk_size", [
    (5),
    (128),
    (8129),
])
async def test_read_common_file(chunk_size):
    """Test on common file"""
    reader = FileReader(file1, sep=b"\n")
    lines = list()
    async for line, offset in reader.get_line(chunk_size=chunk_size):
        lines.append(line)
    assert lines[-1] is None
    lines = lines[:-1]
    with open(file1, "rb") as f:
        lines_orig = f.read().split(b"\n")
        if lines_orig[-1] == b"":
            lines_orig = lines_orig[:-1]
            # I'm doing like this because
            # I need to collect only strings with "\n" symbol at the end

    for i in range(len(lines)):
        assert lines[i] == lines_orig[i]

    assert lines == lines_orig


@pytest.mark.asyncio
@pytest.mark.parametrize("chunk_size", [
    (5),
    (128),
    (8129),
])
async def test_read_file_with_zeroes(chunk_size):
    """Test on file with zeroes at the
    end like MetaTrader Server main log file"""
    reader = FileReader(file2, sep=b"\n")
    lines = list()
    async for line, offset in reader.get_line(chunk_size=chunk_size):
        lines.append(line)
    assert lines[-1] is None
    lines = lines[:-1]
    with open(file2, "rb") as f:
        lines_orig = f.read().split(b"\n")
        if 0 == lines_orig[-1][-1]:
            lines_orig = lines_orig[:-1]
            # I'm doing like this because
            # I need to collect only strings with "\n" symbol at the end
            # and we don't need any \x00
    for i in range(len(lines)):
        assert lines[i] == lines_orig[i]
    assert len(lines) == len(lines_orig)
    assert lines == lines_orig


@pytest.mark.asyncio
@pytest.mark.parametrize("chunk_size", [
    (5),
    (128),
    (8129),
])
async def test_offsets(chunk_size):
    """Test check offsets"""
    reader = FileReader(file1, sep=b"\n")
    offsets = list()
    async for line, offset in reader.get_line(chunk_size=chunk_size):
        if line is not None:
            offsets.append(offset)

    with open(file1, "rb") as f:
        file_orig = f.read()
    offsets_orig = list(
        [offset + 1 for offset, ch in enumerate(file_orig)
         if bytes([ch]) == b'\n']
    )
    assert offsets == offsets_orig



import asyncio
import sys


async def handle_echo(reader, writer):
    # addr = writer.get_extra_info('peername')
    line = " "
    while line:
        line = await reader.readline()
        sys.stdout.write(line.decode())
        break
    writer.close()


loop = asyncio.get_event_loop()
coro = asyncio.start_server(handle_echo, '127.0.0.1', 5045, loop=loop)
server = loop.run_until_complete(coro)


try:
    loop.run_forever()
except KeyboardInterrupt:
    pass


server.close()
loop.run_until_complete(server.wait_closed())
loop.close()



class State:
    def __init__(self, loop, need_shutdown=False, ):
        self._need_shutdown = need_shutdown
        self.loop = loop

    @property
    def need_shutdown(self):
        return self._need_shutdown

    def shutdown(self):
        self._need_shutdown = True

    def __repr__(self,):
        return "shutdown is{} needed".format(
            "" if self._need_shutdown else " not")



import sqlite3


class DataBase():
    def __init__(self, db_file):
        self.db_file = db_file
        self.conn = sqlite3.connect(self.db_file)
        self.c = self.conn.cursor()
        self.create_table_if_not_exists()

    def create_table_if_not_exists(self):
        self.c.execute("""CREATE TABLE IF NOT EXISTS files
                  (filename TEXT PRIMARY KEY,
                  mtime REAL, offset INTEGER(8))""")
        self.conn.commit()

    def insert_ignore_file(self, filename, mtime):
        self.c.execute(
            """INSERT OR IGNORE INTO files VALUES(?,?,0)""",
            (filename, mtime))
        self.conn.commit()

    def get_file(self, filename):
        self.c.execute("""SELECT * FROM files where filename=?""", (filename,))
        result = self.c.fetchone()

        return result

    def update_file(self, filename, offset, mtime):
        self.c.execute("""UPDATE files
                  SET offset=?, mtime=?
                  WHERE filename=?""", (offset, mtime, filename))

        self.conn.commit()

    def close(self):
        self.conn.commit()
        self.conn.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



import pytest
import asyncio
from lsshipper.common.state import State
from lsshipper.connection import logstash_connection


@pytest.mark.asyncio
async def test_read_common_file(event_loop, unused_tcp_port):
    """transfer without disconnect"""
    test = list()
    state = State(event_loop)
    done = asyncio.Event()

    async def handler(reader, writer):
        while True:
            try:
                line = await reader.readline()
            except:
                break
            if not line or line == b'':
                break
            test.append(line.decode())
        done.set()
        writer.close()

    port = unused_tcp_port
    config = {}
    config['connection'] = {}
    config['ssl'] = {}
    config['connection']['host'] = '127.0.0.1'
    config['connection']['port'] = port
    config['ssl']['enable'] = False
    queue = asyncio.Queue(loop=event_loop)
    test_messages = list()

    for i in range(100):
        m = "Message {}\n".format(i)
        test_messages.append(m)
        await queue.put(m)

    server = await asyncio.start_server(
        handler, host='127.0.0.1', port=port, loop=event_loop)
    client = asyncio.ensure_future(logstash_connection(
        queue, state, event_loop, config), )
    await queue.join()
    state.shutdown()
    await done.wait()
    await client
    assert test == test_messages
    server.close()
    await server.wait_closed()


counter = 0


@pytest.mark.asyncio
@pytest.mark.xfail(
    reason="Cant make shure that all messages will be delivered")
async def test_read_common_file_with_disconnects(event_loop, unused_tcp_port):
    """transfer with disconnects"""
    test = list()
    state = State(event_loop)
    global counter
    done = asyncio.Event()
    counter = 0

    async def handler(reader, writer):
        global counter
        while True:
            counter += 1
            line = await reader.readline()
            if not line or line == b'':
                break

            test.append(line.decode())
            if counter % 7 == 0:
                writer.close()
                return
        done.set()
        writer.close()

    port = unused_tcp_port
    config['connection'] = {}
    config['ssl'] = {}
    config['connection']['host'] = '127.0.0.1'
    config['connection']['port'] = port
    config['ssl']['enable'] = False
    queue = asyncio.Queue(loop=event_loop)
    test_messages = list()

    for i in range(100):
        m = "Message {}\n".format(i)
        test_messages.append(m)
        await queue.put(m)

    server = await asyncio.start_server(
        handler, host='127.0.0.1', port=port, loop=event_loop)
    client = asyncio.ensure_future(logstash_connection(
        queue, state, loop=event_loop))
    await queue.join()
    state.shutdown()
    await done.wait()
    await client
    assert len(test) == len(test_messages)
    server.close()
    await server.wait_closed()



import asyncio
from .connection import logstash_connection
from .common.utils import get_files_to_update
import logging
logger = logging.getLogger(name="general")


class FileHandler(object):
    def __init__(self, loop, state, config):
        self.files_in_work = set()
        self.loop = loop
        self.queue = asyncio.Queue(maxsize=10, loop=self.loop)
        self.tasks = list()
        self.state = state
        self.config = config

    async def ship(self, f):
        async for line, offset in f.get_line():

            if self.state.need_shutdown:
                f.sync_to_db(mtime_update=False)
                try:
                    await asyncio.wait_for(self.queue.join(), timeout=10)
                    logger.warning("waiting for deliver all message")
                except:
                    logger.error("not all message was delivered")
                break
            if line:
                message = f.line_to_json(line, offset)
                while not self.state.need_shutdown:
                    try:
                        await asyncio.wait_for(
                            self.queue.put(message),
                            timeout=2)
                        break
                    except asyncio.TimeoutError:
                        pass
            if line is None:  # if line is None we got last line
                logger.debug(
                    "file reading is finished, file: {}".format(f.name))
                f.sync_to_db(mtime_update=True)

        self.files_in_work.remove(f.name)

    async def start(self):
        conn = asyncio.ensure_future(logstash_connection(
            queue=self.queue, state=self.state,
            loop=self.loop, config=self.config))
        while not self.state.need_shutdown:
            files = await get_files_to_update(
                self.loop, self.config)
            for f in files:
                if f.name in self.files_in_work:
                    continue
                f.sync_from_db()
                if not f.need_update:
                    continue
                logger.debug("working with file:{}".format(f.name))
                self.tasks.append(asyncio.ensure_future(self.ship(f)))
                self.files_in_work.add(f.name)
            self.tasks = list(task for task in self.tasks if not task.done())
            await asyncio.sleep(3.14)
            logger.info("queue size: {}".format(self.queue.qsize()))

        while self.tasks:
            self.tasks = list(task for task in self.tasks if not task.done())
            logger.info("stopping tasks, still {}".format(len(self.tasks)))
            await asyncio.sleep(0.3)
        await conn



from configobj import ConfigObj
import logging
import configobj
import re
import sys
import ast
import os
import argparse


logger = logging.getLogger("general")


def prepare_config():
    parser = argparse.ArgumentParser(description='Logstash file shipper')
    parser.add_argument('--config_dir', default='.')
    args = parser.parse_args()
    config_dir = os.path.abspath(args.config_dir)

    log_config_file = os.path.join(config_dir, 'log_config.ini')
    config_file = os.path.join(config_dir, 'config.ini')
    logging.config.fileConfig(log_config_file)

    try:
        _config = ConfigObj(config_file, raise_errors=True)
    except (configobj.ParseError, configobj.UnreprError) as e:
        logger.error(e)
        sys.exit(1)

    config = _config.dict()
    config['files']['pattern'] = ast.literal_eval(config['files']['pattern'])
    config['files']['newline'] = ast.literal_eval(config['files']['newline'])

    config['files']['pattern'] = re.compile(config['files']['pattern'])
    config['files']['newline'] = config['files']['newline'].encode()

    config['ssl']['enable'] = _config['ssl'].as_bool('enable')
    return config



import os
from setuptools import setup, find_packages
import pypandoc


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def parse_requirements(requirements):
    with open(requirements) as f:
        return [l.strip('\n')
                for l in f if l.strip('\n') and not l.startswith('#')]


def long_description():
    return pypandoc.convert('README.md', 'rst').replace("\r", "")


setup(
    name="lsshipper",
    version="0.1.2",
    author="Anton Stavinsky",
    author_email="stavinsky@gmail.com",
    description=("Very basic python logstash shipper"),
    license="MIT",
    keywords="logstash filebeat",
    url="https://github.com/stavinsky/lsshipper/",
    packages=find_packages(exclude=['tests*']),
    tests_require=['pytest', 'pytest-asyncio', ],
    setup_requires=['pytest-runner', ],
    long_description=long_description(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=parse_requirements("requirements.txt"),
    entry_points={
        'console_scripts': [
            'lsshipper = lsshipper.start:main'
            ]
        },
    python_requires='>=3.5',
    data_files=[
        ('', ['README.md'])],
)



import os
from lsshipper.logfile import LogFile
from concurrent.futures import ThreadPoolExecutor
from functools import partial


def get_files_list(config):
    files_conf = config['files']
    pattern = files_conf['pattern']
    dir_path = files_conf["dir_path"]
    files = list()
    for name in os.listdir(dir_path):
        if pattern.match(name):
            full_path = os.path.join(dir_path, name)
            f = LogFile(
                full_path,
                mtime=os.stat(full_path).st_mtime,
                config=config)
            files.append(f)

    return files


async def get_files_to_update(loop, config):

    with ThreadPoolExecutor() as executor:
        files = await loop.run_in_executor(
            executor, partial(get_files_list, config))
    return files



import json
from .reader_aio import FileReader
from .database import DataBase
import os


class LogFile(FileReader):
    def __init__(self, name, mtime=0, offset=0, config={}):
        super().__init__(name, offset=offset)
        self.name = name
        self.mtime = mtime  # Current modify time from db
        self.last_mtime = 0   # Modify time from os.stat
        self.offset = offset
        self.db_file = config['database']["file"]
        self.sep = config['files']['newline']

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = os.path.abspath(name)

    @property
    def need_update(self):
        return self.mtime > self.last_mtime

    def sync_from_db(self,):
        with DataBase(self.db_file) as db:
            db.insert_ignore_file(self.name, 0)
            f = db.get_file(self.name)
        self.offset = f[2]
        self.last_mtime = f[1]

    def sync_to_db(self, mtime_update=False):
        if mtime_update:
            self.last_mtime = self.mtime
        with DataBase(self.db_file) as db:
            db.update_file(self.name, self.offset, self.last_mtime)

    def line_to_json(self, line, offset):
        data = {
            "message": line.decode('utf-8', 'replace'),
            "fields": {
                "program": 'mt4'
            },
            "source": self.name,
            "offset": offset,
        }
        return json.dumps(data) + '\n'

    def __repr__(self):
        return "<LogFile: {}, {}, {}>".format(
            self.name, self.mtime, self.offset)
